/*
 * eth_ecat_pdo_fmmu.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_

#include "ethernet_industry/eth_ecat_buffer.hpp"

namespace free_rtos {

namespace ecat_pdo_fmmu {

class EcatPdoFMMUSlave {
public:
    EcatPdoFMMUSlave(ecat_buffer::EcatBufferSlave& buffer_slave)
    : buffer_slave_(buffer_slave) { }

    template<typename DataType>
    datagram::EcatDatagram<command::LWR, DataType> make_datagram_write(DataType& data) {
        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_write();
        address::Logical logical = fmmu_properties.address;

        return datagram::EcatDatagram<command::LWR, DataType>{ {{logical}}, data };
    }

    template<typename DataType>
    datagram::EcatDatagram<command::LRD, DataType> make_datagram_read(DataType& data) {
        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_read();
        address::Logical logical = fmmu_properties.address;

        return datagram::EcatDatagram<command::LRD, DataType>{ {{logical}}, data };
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;
};

class EthEcatPdoFMMU {
public:
    EthEcatPdoFMMU(ecat_buffer::EthEcatBuffer& ecat_mailbox): ecat_buffer_{ecat_mailbox} { }

    void init();
    void process();

    template<typename DataType, typename... DataTypes>
    void write(DataType& head, DataTypes&... tail) {
        if(sizeof...(tail) + 1 > pdo_fmmu_slaves_.size()) {
            return;
        }

        size_t slave_index = 0;
        auto datagram = pdo_fmmu_slaves_[slave_index].make_datagram_write(head);
        // Сборка и отправка телеграммы с "головы"
        write_recursion(slave_index + 1, datagram, datagram, tail...);
    }

    template<typename DataType, typename... DataTypes>
    void read(DataType& head, DataTypes&... tail) {
        if(sizeof...(tail) + 1 > pdo_fmmu_slaves_.size()) {
            return;
        }

        size_t slave_index = 0;
        auto datagram = pdo_fmmu_slaves_[slave_index].make_datagram_read(head);
        // Сборка и отправка телеграммы с "головы"
        read_recursion(slave_index + 1, datagram, datagram, tail...);
    }

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatPdoFMMUSlave> pdo_fmmu_slaves_;

    template<typename DatagramTypeFirst, typename DatagramTypePrevious, typename... DataTypes>
    void write_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous, DataTypes&... data);

    template<typename DatagramTypeFirst, typename DatagramTypePrevious, typename DataType, typename... DataTypes>
    void write_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous, DataType& head, DataTypes&... tail) {
        auto datagram = pdo_fmmu_slaves_[slave_index].make_datagram_write(head);

        write_recursion(slave_index + 1, first, previous + datagram, tail...);
    }

    template<typename DatagramTypeFirst, typename DatagramTypePrevious>
    void write_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous) {
        datagram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();

        do {
            telegram.transfer(first);
        } while(first.get_all_wkc() < slave_index);
    }


    template<typename DatagramTypeFirst, typename DatagramTypePrevious, typename... DataTypes>
    void read_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous, DataTypes&... data);

    template<typename DatagramTypeFirst, typename DatagramTypePrevious, typename DataType, typename... DataTypes>
    void read_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous, DataType& head, DataTypes&... tail) {
        auto datagram = pdo_fmmu_slaves_[slave_index].make_datagram_read(head);

        read_recursion(slave_index + 1, first, previous + datagram, tail...);
    }

    template<typename DatagramTypeFirst, typename DatagramTypePrevious>
    void read_recursion(size_t slave_index, DatagramTypeFirst& first, DatagramTypePrevious& previous) {
        datagram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();

        do {
            telegram.transfer(first);
        } while(first.get_all_wkc() < slave_index);
    }

    void wait_op();
};

} // namespace ecat_pdo_fmmu

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_ */
